/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description
    Reads .msh file as written by Gmsh.

    Needs surface elements on mesh to be present and aligned with outside faces
    of the mesh. I.e. if the mesh is hexes, the outside faces need to be quads

    Note: There is something seriously wrong with the ordering written in the
    .msh file. Normal operation is to use the ordering as described
    in the manual. Use the -autoInvert to invert based on the geometry.
    Not very well tested.


Several modification were made relative to the gmshToFoam utility shipped with OpenFOAM v1.2,
neither of which have been tested extensively.  


    1)  primitivePatch::meshPointMap() call was moved outside of findFace() instead
        const Map<label> meshPointMap is passed as an argument.  This was done so that
        the primitivePatch::meshPointMap() call could be moved outside the loop which was
        going over all the "faces" of all the "patches" in the domain.  This should improve
        the conversion speed.
          
        
    2)  The code now uses the element (cell,face) physical region id number to create
        cell zones and faces zones (similar to fluentMeshWithInternalFaces).

        "-noCellZones" and "-noFaceZones" represses the addition of this information to the mesh
        
        A use of the cell zone information, is for field initialization with the "setFields" utility.
        see the classes:  topoSetSource, zoneToCell.  

         
    3)  The addition of the flag "-verbose <int>" to control the amount of screen output, verbose = 0,
        produce similar output to OpenFoam v1.2 gmshToFoam converter.


\*---------------------------------------------------------------------------*/

#include "argList.H"
#include "Time.H"
#include "polyMesh.H"
#include "IFstream.H"
#include "wallPolyPatch.H"
#include "cellModeller.H"
#include "repatchPolyMesh.H"

using namespace Foam;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

// Element type numbers
static label MSHTRI   = 2;
static label MSHQUAD  = 3;
static label MSHTET   = 4;
static label MSHPYR   = 7;
static label MSHPRISM = 6;
static label MSHHEX   = 5;


void renumber
(
    const Map<label>& mshToFoam,
    labelList& labels
)
{
    forAll(labels, labelI)
    {
        labels[labelI] = mshToFoam[labels[labelI]];
    }
}


/*
    finds and returns the face, "faceI" in the OpenFOAM polyMesh "mesh" which 
    have the same the vertices as the gmesh face "meshF"

    a list of candidate faces is first generate from the "mesh" point->face mapping ("pointFaces()")

    if no match is found return -1
*/

label findInternalFace( const polyMesh &mesh, const labelList& meshF, const int verbose )
{
    
    labelList candidateFaces = mesh.pointFaces()[meshF[0]];

    forAll( candidateFaces, i) 
    {
        label faceI = candidateFaces[i];
        label nMatched = 0;
        const face& f = mesh.allFaces()[faceI];
    
        forAll(f, fp)
        {
            if (findIndex(meshF, f[fp]) != -1)
            {
                nMatched++;
            }
        }

        if (nMatched == meshF.size())
        {
            return faceI;
        }
    }

   
    if (verbose > 0)
    {
    Warning << "Could not match gmsh face " << meshF
        << " to any of the interior mesh faces that share the same zeroth point:"
        << candidateFaces << endl;
    }

    return -1;
}

// Find face in pp which uses all vertices in meshF (in mesh point labels)

label findFace(const primitivePatch& pp, const labelList& meshF, const Map<label> &meshPointMap, const int verbose)
{
    // move creation of map to outer loop
    // const Map<label> meshPointMap = pp.meshPointMap();

    // meshF[0] in pp labels.

    if (!meshPointMap.found(meshF[0]))
    {
        Warning<< "Not using gmsh face " << meshF
            << " since zero vertex is not on boundary of polyMesh" << endl;


        return -1;
    }

    // Find faces using first point
    const labelList& pFaces = pp.pointFaces()[meshPointMap[meshF[0]]];

    // Go through all these faces and check if there is one which uses all of
    // meshF vertices (in any order ;-)
    forAll(pFaces, i)
    {
        label faceI = pFaces[i];

        const face& f = pp[faceI];

        // Count uses of vertices of meshF for f
        label nMatched = 0;

        forAll(f, fp)
        {
            if (findIndex(meshF, f[fp]) != -1)
            {
                nMatched++;
            }
        }

        if (nMatched == meshF.size())
        {
            return faceI;
        }
    }

    if (verbose > 0)
    {
    Warning << "Could not match gmsh face " << meshF
        << " to any of the outside mesh faces that share the same zeroth point:"
        << IndirectList<face>(pp, pFaces) << endl;
    }


    return -1;
}


// Determine whether cell is inside-out by checking for any wrong-oriented
// face.
bool correctOrientation(const pointField& points, const cellShape& shape)
{
    // Get centre of shape.
    point cc(shape.centre(points));

    // Get outwards pointing faces.
    faceList faces(shape.faces());

    forAll(faces, i)
    {
        const face& f = faces[i];

        vector n(f.normal(points));

        // Check if vector from any point on face to cc points outwards
        if (((points[f[0]] - cc) & n) < 0)
        {
            // Incorrectly oriented
            return false;
        }
    }

    return true;
}


// Main program:

int main(int argc, char *argv[])
{
    argList::noParallel();
    argList::validArgs.append(".msh file");
    argList::validOptions.insert("autoInvert", "");

    argList::validOptions.insert("noCellZones", "");
    argList::validOptions.insert("verbose", "label");
    argList::validOptions.insert("noFaceZones","");

#   include "setRootCase.H"
#   include "createTime.H"

    fileName mshName(args.args()[3]);

    bool autoInvert = args.options().found("autoInvert");
    bool noCellZones = args.options().found("noFaceZones");
    bool noFaceZones = args.options().found("noCellZones");

    label verbose = 0;    
    if ( args.options().found("verbose") ) {
        verbose = atoi( args.options()["verbose"].c_str() );
    }
    Info << "verbose = " << verbose << endl;

    IFstream inFile(mshName);

    word tag(inFile);

    if (tag != "$NOD")
    {
        FatalErrorIn(args.executable())
            << "Did not find $NOD tag on line " << inFile.lineNumber()
            << exit(FatalError);
    }

    label nVerts;

    inFile >> nVerts;

    Info<< "Read nVerts:" << nVerts << endl << endl;

    pointField points(nVerts);

    Map<label> mshToFoam(nVerts);

    for (label pointI = 0; pointI < nVerts; pointI++)
    {
        label mshLabel;
        scalar xVal, yVal, zVal;

        inFile >> mshLabel >> xVal >> yVal >> zVal;

        point& pt = points[pointI];

        pt.x() = xVal;
        pt.y() = yVal;
        pt.z() = zVal;

        mshToFoam.insert(mshLabel, pointI);
    }
    inFile >> tag;

    if (tag != "$ENDNOD")
    {
        FatalErrorIn(args.executable())
            << "Did not find $ENDNOD tag on line " << inFile.lineNumber()
            << exit(FatalError);
    }

    inFile >> tag;

    if (tag != "$ELM")
    {
        FatalErrorIn(args.executable())
            << "Did not find $ELM tag on line " << inFile.lineNumber()
            << exit(FatalError);
    }

    label nElems;

    inFile >> nElems;

    Info<< "Read nElems:" << nElems << endl << endl;


    // Storage for all cells. Too big. Shrink later
    cellShapeList cells(nElems);

    label cellI = 0;

    const cellModel& hex = *(cellModeller::lookup("hex"));
    const cellModel& prism = *(cellModeller::lookup("prism"));
    const cellModel& pyr = *(cellModeller::lookup("pyr"));
    const cellModel& tet = *(cellModeller::lookup("tet"));

    face triPoints(3);
    face quadPoints(4);
    labelList tetPoints(4);
    labelList pyrPoints(5);
    labelList prismPoints(6);
    labelList hexPoints(8);

    label nTet = 0;
    label nPyr = 0;
    label nPrism = 0;
    label nHex = 0;


    // From gmsh physical region to Foam patch
    Map<label> regionToPatch;


    // add support of different regions
    Map<label> volumeRegion;
    List<List<label> > zoneCells(0);

    // Storage for patch faces.
    List<DynamicList<face> > patchFaces(0);
   
    int iPrint = nElems;
    if (verbose > 0) iPrint =  nElems/verbose;

    for (label elemI = 0; elemI < nElems; elemI++)
    {
        label elmNumber, elmType, regPhys, regElem, nNodes;
        //label zoneI;
        inFile >> elmNumber >> elmType >> regPhys >> regElem >> nNodes;

        // regPhys on surface elements is region number.

        labelList verts(nNodes);

        
        if (elemI % iPrint == 0 && verbose > 0)
             Info  << "Reading element " << elemI << " of " << nElems << endl;  

        if (elmType == MSHTET || elmType == MSHPYR || elmType == MSHPRISM || elmType == MSHHEX) 
        {

            Map<label>::iterator regFnd = volumeRegion.find(regPhys);
            label zoneI = -1;
            if ( regFnd == volumeRegion.end() ) 
            {
                zoneCells.setSize( zoneCells.size() + 1);
                zoneI = zoneCells.size() - 1;
                Info << "Mapping volume region " << regPhys << " to Foam region " << zoneI << endl;
                volumeRegion.insert( regPhys, zoneI );
            }
            else
            {
                zoneI = regFnd();
            }
            if (zoneI > -1) 
            {
                zoneCells[zoneI].setSize( zoneCells[zoneI].size() + 1);
                zoneCells[zoneI][ zoneCells[zoneI].size() - 1] = cellI;
            }
            else
            {
                Warning << "Zone index = " << zoneI << endl;
            }
        }

        if (elmType == MSHTRI)
        {
            inFile >> triPoints[0] >> triPoints[1] >> triPoints[2];

            renumber(mshToFoam, triPoints);

            Map<label>::iterator regFnd = regionToPatch.find(regPhys);

            label patchI = -1;
            if (regFnd == regionToPatch.end())
            {
                // New region. Allocate patch for it.
                patchFaces.setSize(patchFaces.size() + 1);

                patchI = patchFaces.size()-1;

                Info<< "Mapping region " << regPhys << " to Foam patch "
                    << patchI << endl;
                regionToPatch.insert(regPhys, patchI);
            }
            else
            {
                // Existing patch for region
                patchI = regFnd();
            }

            // Add triangle to correct patchFaces.
            patchFaces[patchI].append(triPoints);
        }
        else if (elmType == MSHQUAD)
        {
            inFile
                >> quadPoints[0] >> quadPoints[1] >> quadPoints[2]
                >> quadPoints[3];

            renumber(mshToFoam, quadPoints);

            Map<label>::iterator regFnd = regionToPatch.find(regPhys);

            label patchI = -1;
            if (regFnd == regionToPatch.end())
            {
                // New region. Allocate patch for it.
                patchFaces.setSize(patchFaces.size() + 1);

                patchI = patchFaces.size()-1;

                Info<< "Mapping region " << regPhys << " to Foam patch "
                    << patchI << endl;
                regionToPatch.insert(regPhys, patchI);
            }
            else
            {
                // Existing patch for region
                patchI = regFnd();
            }

            // Add quad to correct patchFaces.
            patchFaces[patchI].append(quadPoints);
        }
        else if (elmType == MSHTET)
        {
   
            inFile
                >> tetPoints[0] >> tetPoints[1] >> tetPoints[2]
                >> tetPoints[3];

            renumber(mshToFoam, tetPoints);

            cells[cellI++] = cellShape(tet, tetPoints);

            nTet++;
        }
        else if (elmType == MSHPYR)
        {
            inFile
                >> pyrPoints[0] >> pyrPoints[1] >> pyrPoints[2]
                >> pyrPoints[3] >> pyrPoints[4];

            renumber(mshToFoam, pyrPoints);

            cells[cellI] = cellShape(pyr, pyrPoints);
            cellI++;
            nPyr++;
        }
        else if (elmType == MSHPRISM)
        {
            inFile
                >> prismPoints[0] >> prismPoints[1] >> prismPoints[2]
                >> prismPoints[3] >> prismPoints[4] >> prismPoints[5];

            renumber(mshToFoam, prismPoints);

            cells[cellI] = cellShape(prism, prismPoints);

            const cellShape& cell = cells[cellI];

            if (autoInvert && !correctOrientation(points, cell))
            {
                Info<< "Inverting prism " << cellI << endl;
                // Reorder prism.
                prismPoints[0] = cell[0];
                prismPoints[1] = cell[2];
                prismPoints[2] = cell[1];
                prismPoints[3] = cell[3];
                prismPoints[4] = cell[4];
                prismPoints[5] = cell[5];

                cells[cellI] = cellShape(prism, prismPoints);
            }

            cellI++;

            nPrism++;
        }
        else if (elmType == MSHHEX)
        {
            inFile
                >> hexPoints[0] >> hexPoints[1]
                >> hexPoints[2] >> hexPoints[3]
                >> hexPoints[4] >> hexPoints[5]
                >> hexPoints[6] >> hexPoints[7];

            renumber(mshToFoam, hexPoints);

            cells[cellI] = cellShape(hex, hexPoints);

            const cellShape& cell = cells[cellI];

            if (autoInvert && !correctOrientation(points, cell))
            {
                Info<< "Inverting hex " << cellI << endl;

                // Reorder hex.
                hexPoints[0] = cell[4];
                hexPoints[1] = cell[5];
                hexPoints[2] = cell[6];
                hexPoints[3] = cell[7];
                hexPoints[4] = cell[0];
                hexPoints[5] = cell[1];
                hexPoints[6] = cell[2];
                hexPoints[7] = cell[3];

                cells[cellI] = cellShape(hex, hexPoints);
            }

            cellI++;

            nHex++;
        }
        else
        {
            // Unhandled element. Read the vertices.
            for (label i = 0; i < nNodes; i++)
            {
                label mshLabel;

                inFile >> mshLabel;
            }
        }


        
    }
    cells.setSize(cellI);

    forAll(patchFaces, patchI)
    {
        patchFaces[patchI].shrink();
    }


    Info<< "Cells:" << endl
        << "    total:" << cells.size() << endl
        << "    hex  :" << nHex << endl
        << "    prism:" << nPrism << endl
        << "    pyr  :" << nPyr << endl
        << "    tet  :" << nTet << endl
        << endl;


    Info<< "Patches:" << nl
        << "Patch\tSize" << endl;

    forAll(patchFaces, patchI)
    {
        Info<< "    " << patchI << '\t' << patchFaces[patchI].size() << endl;
    }
    Info<< endl;


    // Problem is that the orientation of the patchFaces does not have to
    // be consistent with the outwards orientation of the mesh faces. So
    // we have to construct the mesh in two stages:
    // 1. define mesh with all boundary faces in one patch
    // 2. use the read patchFaces to find the corresponding boundary face
    //    and repatch it.



    // Create correct number of patches
    // (but without any faces in it)
    faceListList boundaryFaces(patchFaces.size());

    wordList boundaryPatchNames(boundaryFaces.size());

    forAll(boundaryPatchNames, patchI)
    {
        boundaryPatchNames[patchI] = word("patch") + name(patchI);
    }

    wordList boundaryPatchTypes(boundaryFaces.size(), polyPatch::typeName);
    word defaultFacesType = polyPatch::typeName;
    wordList boundaryPatchPhysicalTypes
    (
        boundaryFaces.size(),
        polyPatch::typeName
    );

    if (verbose > 0) Info<< "Start:Repatch Poly Mesh\n" << endl;

    repatchPolyMesh mesh
    (
        IOobject
        (
            polyMesh::defaultRegion,
            runTime.constant(),
            runTime
        ),
        points,
        cells,
        boundaryFaces,
        boundaryPatchNames,
        boundaryPatchTypes,
        defaultFacesType,
        boundaryPatchPhysicalTypes
    );

    if (verbose > 0) Info<< "End:Repatch Poly Mesh\n" << endl;

    if (verbose > 0) 
    {
        Info << "size of boundary mesh " << mesh.boundaryMesh().size() << endl;
        forAll( mesh.boundaryMesh(), i)
        {
            Info << "\t" << i << " " << mesh.boundaryMesh()[i].size() << endl;
        }
    }

    // Now use the patchFaces to patch up the outside faces of the mesh.
    // Get the patch for all the outside faces (= default patch added as last)
    const polyPatch& pp = mesh.boundaryMesh()[mesh.boundaryMesh().size()-1];
    const Map<label> boundaryMeshPointMap = pp.meshPointMap();
  
    // storage for indicies of faces belonging to 
    List<List<label> > zoneFaces( patchFaces.size() );

    // Go through all the patchFaces and find corresponding face in pp.
     forAll(patchFaces, patchI)
    {
        const DynamicList<face>& pFaces = patchFaces[patchI];

        Info  << "Finding faces of patch " << patchI << endl; 

        iPrint = pFaces.size();
        if (verbose > 0) iPrint = iPrint/verbose;
        
        forAll(pFaces, i)
        {
            const face& f = pFaces[i];
            if (verbose > 0 && ((i % iPrint) == 0))
                  Info  << "\tFinding face " << i << " of " << pFaces.size() << endl;  

            // Find face in pp using all vertices of f.
            //label patchFaceI = findFace(pp, f);
            label patchFaceI = findFace(pp, f, boundaryMeshPointMap, verbose);

            if (patchFaceI != -1)
            {
                label meshFaceI = pp.start() + patchFaceI;
                mesh.changePatchID(meshFaceI, patchI);
            }           
            else
            {  
                // gmesh face was not found on the exterior, look in the interior
                patchFaceI = findInternalFace(mesh, f, verbose);
                if (patchFaceI != -1) 
                {
                    label k = zoneFaces[patchI].size();
                    zoneFaces[patchI].setSize( k + 1 );
                    zoneFaces[patchI][ k ] = patchFaceI;
                }
                else
                {
                    Warning << "Could not match gmsh face " << f
                        << " to any of the interior or exterior mesh faces that share the same zeroth point:" << endl;
                }
   
            }
            
            
        }
        
    }
  
    mesh.repatch();
    Info << "repatch" << endl;


/*


    Building zone info - adapted from fluentMeshToFoamWithInternals


*/
    Info << "Building zone info" << endl;


    Info << "Building point zones" << endl;
    List<pointZone*> pz;

    /* dummy point zone is not needed
    pz[0]=new pointZone
	  (
	   word("dummy_points"),
	   labelList(0),
	   998,
	   mesh.pointZones()
	   );
    */

/*    

    Build Cells Zones

*/    
    
    List<cellZone*> cz;

    if (noCellZones) {
        Info << "cell Zone info not built" << endl;
    }
    else
    {
        cz.setSize( zoneCells.size() );
        for (label zoneI = 0; zoneI < zoneCells.size(); zoneI++)
        {
            std::ostringstream buffer;
            buffer << zoneI;
            string name = "zone_" + buffer.str();
            Info << "Building zone info for " << name << ", " << zoneI << " of " << zoneCells.size() <<
            " with " << zoneCells[zoneI].size() << " cells, min = " << min(zoneCells[zoneI]) 
                    << ", max = " << max(zoneCells[zoneI]) << endl;
            cz[zoneI] = new cellZone
	            (
	            name,
	            zoneCells[zoneI],
	            zoneI,
	       mesh.cellZones()
	       );
        }    
    }

/* 

    Build Face Zones    

*/

    int nFaceZones = 0;
    for (label zoneI = 0; zoneI < zoneFaces.size(); zoneI ++)
    {
        if ( zoneFaces[zoneI].size() > 0 ) 
        {
        nFaceZones += 1;    
        }
    }

    Info << "Mesh has " << nFaceZones << " internal face zones" << endl;
    List<faceZone*> fz;
     
    /* need to have a least one face zone, to make cell zones output */
    

    if (noFaceZones) {
        Info << "face Zone info not built" << endl;
    }
    else
    {
        fz.setSize(nFaceZones);
        label j = 0;
        for (label zoneI = 0; zoneI < zoneFaces.size(); zoneI++)
        {

            if (zoneFaces[zoneI].size() > 0 ) 
            {

                std::ostringstream buffer;
                buffer << zoneI;
                string name = "faceZone_" + buffer.str();  
                Info << "Building face zone info for " << name << ", " << zoneI << " of " << zoneFaces.size()
                << " with " << zoneFaces[zoneI].size() << " faces, min = " <<  min(zoneFaces[zoneI]) 
                    << ", max = " << max(zoneFaces[zoneI]) << endl;

                fz[j] = new faceZone
	                (
	                name,
	                zoneFaces[zoneI],
                    boolList(zoneFaces[zoneI].size(),true),
	                zoneI,
	                mesh.faceZones()
	                );
                j++;
            }
        }

    }

    // add zones to mesh


    if ( (!noFaceZones) || (!noCellZones) ) { 


        // need a dummy face zone to force the writing of cellZones
        if (nFaceZones < 1 && (!noCellZones) )
        {
            fz.setSize(1);
            fz[0]=new faceZone
	        (
	            word("dummy"),
	            labelList(0),
	            boolList(0,false),
	            999,
	            mesh.faceZones()
	        );
        }

        Info << "Adding zones to mesh" << endl;
        mesh.addZones(pz,fz,cz);
    }

    //Get polyMesh to write to constant
    runTime.setTime(instant(runTime.constant()), 0);
    Info << "runTime.setTime" << endl;
 
    Info << "Completed writing zone info" << endl;
    mesh.write();


    Info<< "End\n" << endl;

    return 0;
}


// ************************************************************************* //

